@@ -47,18 +47,19 @@ Paul Licameli split from AudacityProject.cpp
 #include "SelectionState.h"
 #include "Tags.h"
 #include "TempDirectory.h"
+#include "TempoChange.h"
 #include "TimeDisplayMode.h"
 #include "TrackFocus.h"
 #include "TrackPanel.h"
-#include "UndoManager.h"
+#include "UndoTracks.h"
 #include "UserException.h"
+#include "ViewInfo.h"
 #include "WaveClip.h"
 #include "WaveTrack.h"
-#include "WideClip.h"
+#include "WaveTrackUtilities.h"
 #include "XMLFileReader.h"
 #include "import/ImportStreamDialog.h"
 #include "prefs/ImportExportPrefs.h"
-#include "tracks/playabletrack/wavetrack/WaveTrackUtils.h"
 #include "widgets/FileHistory.h"
 #include "widgets/UnwritableLocationErrorDialog.h"
 #include "widgets/Warning.h"
@@ -98,7 +99,7 @@ void ProjectFileManager::DiscardAutosave(const FilePath &filename)
 
    if (projectFileManager.mLastSavedTracks) {
       for (auto wt : projectFileManager.mLastSavedTracks->Any<WaveTrack>())
-         wt->CloseLock();
+         WaveTrackUtilities::CloseLock(*wt);
       projectFileManager.mLastSavedTracks.reset();
    }
 
@@ -190,9 +191,39 @@ auto ProjectFileManager::ReadProjectFile(
       // user selects Save().
       // Do this before FixTracks might delete zero-length clips!
       mLastSavedTracks = TrackList::Create( nullptr );
-      for (auto t : tracks)
-         mLastSavedTracks->Append(
-            move(*t->Duplicate(Track::DuplicateOptions{}.Backup())));
+      WaveTrack *leader{};
+      tracks.Any().Visit(
+         [&](WaveTrack& track) {
+            // A rare place where TrackList::Channels remains necessary, to
+            // visit the right channels of stereo tracks not yet "zipped",
+            // otherwise later, CloseLock() will be missed for some sample
+            // blocks and corrupt the project
+            for (const auto pChannel : TrackList::Channels(&track))
+            {
+               auto left = leader;
+               auto newTrack =
+                  pChannel->Duplicate(Track::DuplicateOptions {}.Backup());
+               leader = left ? nullptr // now visiting the right channel
+                        :
+                        (pChannel->GetLinkType() == Track::LinkType::None) ?
+                               nullptr // now visiting a mono channel
+                               :
+                               static_cast<WaveTrack*>(newTrack.get())
+                  // now visiting a left channel
+                  ;
+               mLastSavedTracks->Add(newTrack);
+               if (left)
+                  // Zip clips allowing misalignment -- this may be a legacy
+                  // project.  This duplicate track will NOT be used for normal
+                  // editing, but only later to visit all the sample blocks that
+                  // existed at last save time.
+                  left->ZipClips(false);
+            }
+         },
+         [this](Track& track) {
+            mLastSavedTracks->Add(
+               track.Duplicate(Track::DuplicateOptions {}.Backup()));
+         });
 
       FixTracks(
          tracks,
@@ -261,7 +292,6 @@ bool ProjectFileManager::Save()
 {
    auto &projectFileIO = ProjectFileIO::Get(mProject);
 
-   
    if (auto action = ProjectFileIOExtensionRegistry::OnSave(
           mProject, [this](auto& path, bool rename)
           { return DoSave(audacity::ToWXString(path), rename); });
@@ -325,25 +355,6 @@ bool ProjectFileManager::DoSave(const FilePath & fileName, const bool fromSaveAs
          return false;
       }
 
-      auto &tracks = TrackList::Get( proj );
-      if (tracks.empty())
-      {
-         if (UndoManager::Get( proj ).UnsavedChanges() &&
-               settings.EmptyCanBeDirty())
-         {
-            int result = AudacityMessageBox(
-               XO(
-   "Your project is now empty.\nIf saved, the project will have no tracks.\n\nTo save any previously open tracks:\nClick 'No', Edit > Undo until all tracks\nare open, then File > Save Project.\n\nSave anyway?"),
-               XO("Warning - Empty Project"),
-               wxYES_NO | wxICON_QUESTION,
-               &window);
-            if (result == wxNO)
-            {
-               return false;
-            }
-         }
-      }
-
       wxULongLong fileSize = wxFileName::GetSize(projectFileIO.GetFileName());
 
       wxDiskspaceSize_t freeSpace;
@@ -416,8 +427,7 @@ bool ProjectFileManager::DoSave(const FilePath & fileName, const bool fromSaveAs
 
    auto &tracks = TrackList::Get(proj);
    for (auto t : tracks)
-      mLastSavedTracks->Append(
-         move(*t->Duplicate(Track::DuplicateOptions{}.Backup())));
+      mLastSavedTracks->Add(t->Duplicate(Track::DuplicateOptions{}.Backup()));
 
    // If we get here, saving the project was successful, so we can DELETE
    // any backup project.
@@ -482,13 +492,6 @@ bool ProjectFileManager::SaveAs(bool allowOverwrite /* = false */)
 
    TranslatableString title = XO("%sSave Project \"%s\" As...")
       .Format( Restorer.sProjNumber, Restorer.sProjName );
-   TranslatableString message = XO("\
-'Save Project' is for an Audacity project, not an audio file.\n\
-For an audio file that will open in other apps, use 'Export'.\n");
-
-   if (ShowWarningDialog(&window, wxT("FirstProjectSave"), message, true) != wxID_OK) {
-      return false;
-   }
 
    bool bPrompt = (project.mBatchMode == 0) || (projectFileIO.GetFileName().empty());
    FilePath fName;
@@ -792,20 +795,22 @@ void ProjectFileManager::CompactProjectOnClose()
    if (mLastSavedTracks)
    {
       for (auto wt : mLastSavedTracks->Any<WaveTrack>())
-         wt->CloseLock();
+         WaveTrackUtilities::CloseLock(*wt);
 
       // Attempt to compact the project
-      projectFileIO.Compact( { mLastSavedTracks.get() } );
+      projectFileIO.Compact({ mLastSavedTracks.get() });
 
-      if ( !projectFileIO.WasCompacted() &&
-          UndoManager::Get( project ).UnsavedChanges() ) {
+      if (
+         !projectFileIO.WasCompacted() &&
+         UndoManager::Get(project).UnsavedChanges())
+      {
          // If compaction failed, we must do some work in case of close
          // without save.  Don't leave the document blob from the last
          // push of undo history, when that undo state may get purged
          // with deletion of some new sample blocks.
          // REVIEW: UpdateSaved() might fail too.  Do we need to test
          // for that and report it?
-         projectFileIO.UpdateSaved( mLastSavedTracks.get() );
+         projectFileIO.UpdateSaved(mLastSavedTracks.get());
       }
    }
 }
@@ -1031,8 +1036,13 @@ void ProjectFileManager::FixTracks(TrackList& tracks,
    const std::function<void(const TranslatableString&)>& onError,
    const std::function<void(const TranslatableString&)>& onUnlink)
 {
-   Track* unlinkedTrack {};
-   for (const auto t : tracks) {
+   // This is successively assigned the left member of each pair that
+   // becomes unlinked
+   Track::Holder unlinkedTrack;
+   // Beware iterator invalidation, because stereo channels get zipped,
+   // replacing WaveTracks
+   for (auto iter = tracks.begin(); iter != tracks.end();) {
+      auto t = (*iter++)->SharedPointer();
       const auto linkType = t->GetLinkType();
       // Note, the next function may have an important upgrading side effect,
       // and return no error; or it may find a real error and repair it, but
@@ -1041,14 +1051,27 @@ void ProjectFileManager::FixTracks(TrackList& tracks,
          onError(XO("A channel of a stereo track was missing."));
          unlinkedTrack = nullptr;
       }
-      if(unlinkedTrack != nullptr)
-      {
+      if (!unlinkedTrack) {
+         if (linkType != ChannelGroup::LinkType::None &&
+            t->NChannels() == 1) {
+            // The track became unlinked.
+            // It should NOT have been replaced with a "zip"
+            assert(t->GetOwner().get() == &tracks);
+            // Wait until LinkConsistencyFix is called on the second track
+            unlinkedTrack = t;
+            // Fix the iterator, which skipped the right channel before the
+            // unlinking
+            iter = tracks.Find(t.get());
+            ++iter;
+         }
+      }
+      else {
          //Not an elegant way to deal with stereo wave track linking
          //compatibility between versions
-         if(const auto left = dynamic_cast<WaveTrack*>(unlinkedTrack))
-         {
-            if(const auto right = dynamic_cast<WaveTrack*>(t))
-            {
+         if (const auto left = dynamic_cast<WaveTrack*>(unlinkedTrack.get())) {
+            if (const auto right = dynamic_cast<WaveTrack*>(t.get())) {
+               // As with the left, it should not have vanished from the list
+               assert(right->GetOwner().get() == &tracks);
                left->SetPan(-1.0f);
                right->SetPan(1.0f);
                RealtimeEffectList::Get(*left).Clear();
@@ -1067,13 +1090,6 @@ void ProjectFileManager::FixTracks(TrackList& tracks,
          unlinkedTrack = nullptr;
       }
 
-      if(linkType != ChannelGroup::LinkType::None &&
-         t->GetLinkType() == ChannelGroup::LinkType::None)
-      {
-         //Wait when LinkConsistencyFix is called on the second track
-         unlinkedTrack = t;
-      }
-
       if (const auto message = t->GetErrorOpening()) {
          wxLogWarning(
             wxT("Track %s had error reading clip values from project file."),
@@ -1141,7 +1157,7 @@ AudacityProject *ProjectFileManager::OpenProjectFile(
       // here is a better way to accomplish the intent, doing like what happens
       // when the project closes:
       for (auto pTrack : tracks.Any<WaveTrack>())
-         pTrack->CloseLock();
+         WaveTrackUtilities::CloseLock(*pTrack);
 
       tracks.Clear(); //tracks.Clear(true);
 
@@ -1183,19 +1199,14 @@ ProjectFileManager::AddImportedTracks(const FilePath &fileName,
       !(tracks.Any<PlayableTrack>() + &PlayableTrack::GetSolo).empty();
    if (projectHasSolo) {
       for (auto &group : newTracks)
-         for (const auto pTrack : group->Any<PlayableTrack>())
+         if (auto pTrack = dynamic_cast<PlayableTrack*>(group.get()))
             pTrack->SetMute(true);
    }
 
-   // Must add all tracks first (before using Track::IsLeader)
    for (auto &group : newTracks) {
-      if (group->empty()) {
-         assert(false);
-         continue;
-      }
-      for (const auto pTrack : group->Any<WaveTrack>())
+      if (auto pTrack = dynamic_cast<WaveTrack*>(group.get()))
          results.push_back(pTrack);
-      tracks.Append(std::move(*group));
+      tracks.Add(group);
    }
    newTracks.clear();
 
@@ -1219,7 +1230,7 @@ ProjectFileManager::AddImportedTracks(const FilePath &fileName,
          if (newRate == 0)
             newRate = wt.GetRate();
          const auto trackName = wt.GetName();
-         for(const auto& interval : wt.Intervals())
+         for (const auto &interval : wt.Intervals())
             interval->SetName(trackName);
       });
    }
@@ -1334,7 +1345,38 @@ private:
 
 bool ProjectFileManager::Import(const FilePath& fileName, bool addToHistory)
 {
-   return Import(std::vector<FilePath> { fileName }, addToHistory);
+   wxArrayString fileNames;
+   fileNames.Add(fileName);
+   return Import(std::move(fileNames), addToHistory);
+}
+
+bool ProjectFileManager::Import(wxArrayString fileNames, bool addToHistory)
+{
+   fileNames.Sort(FileNames::CompareNoCase);
+   if (!ProjectFileManager::Get(mProject).ImportAndRunTempoDetection(
+          std::vector<wxString> { fileNames.begin(), fileNames.end() },
+          addToHistory))
+      return false;
+   // Last track in the project is the one that was just added. Use it for
+   // focus, etc.
+   Track* lastTrack = nullptr;
+   const auto range = TrackList::Get(mProject).Any<Track>();
+   assert(!range.empty());
+   if(range.empty())
+      return false;
+   lastTrack = *(range.rbegin());
+   BasicUI::CallAfter([wTrack = lastTrack->weak_from_this(), wProject = mProject.weak_from_this()] {
+      const auto project = wProject.lock();
+      const auto track = wTrack.lock();
+      if (!project || !track)
+         return;
+      auto& viewPort = Viewport::Get(*project);
+      TrackFocus::Get(*project).Set(track.get(), true);
+      viewPort.ZoomFitHorizontally();
+      viewPort.ShowTrack(*track);
+      viewPort.HandleResize(); // Adjust scrollers for NEW track sizes.
+   });
+   return true;
 }
 
 namespace
@@ -1375,7 +1417,7 @@ std::vector<std::shared_ptr<MIR::AnalyzedAudioClip>> RunTempoDetection(
 }
 } // namespace
 
-bool ProjectFileManager::Import(
+bool ProjectFileManager::ImportAndRunTempoDetection(
    const std::vector<FilePath>& fileNames, bool addToHistory)
 {
    const auto projectWasEmpty =
@@ -1384,12 +1426,15 @@ bool ProjectFileManager::Import(
    const auto success = std::all_of(
       fileNames.begin(), fileNames.end(), [&](const FilePath& fileName) {
          std::shared_ptr<ClipMirAudioReader> resultingReader;
-         const auto success = Import(fileName, addToHistory, resultingReader);
+         const auto success = DoImport(fileName, addToHistory, resultingReader);
          if (success && resultingReader)
             resultingReaders.push_back(std::move(resultingReader));
          return success;
       });
-   if (success && !resultingReaders.empty())
+   // At the moment, one failing import doesn't revert the project state, hence
+   // we still run the analysis on what was successfully imported.
+   // TODO implement reverting of the project state on failure.
+   if (!resultingReaders.empty())
    {
       const auto pProj = mProject.shared_from_this();
       BasicUI::CallAfter([=] {
@@ -1403,7 +1448,7 @@ bool ProjectFileManager::Import(
 }
 
 // If pNewTrackList is passed in non-NULL, it gets filled with the pointers to NEW tracks.
-bool ProjectFileManager::Import(
+bool ProjectFileManager::DoImport(
    const FilePath& fileName, bool addToHistory,
    std::shared_ptr<ClipMirAudioReader>& resultingReader)
 {
@@ -1487,17 +1532,18 @@ bool ProjectFileManager::Import(
          return false;
 
       const auto projectTempo = ProjectTimeSignature::Get(project).GetTempo();
-      for (auto trackList : newTracks)
-         for (auto track : *trackList)
-            track->OnProjectTempoChange(projectTempo);
+      for (auto track : newTracks)
+         DoProjectTempoChange(*track, projectTempo);
 
-      if (!newTracks.empty() && newTracks[0])
+      if (newTracks.size() == 1)
       {
-         const auto waveTracks = (*newTracks[0]).Any<WaveTrack>();
-         if (waveTracks.size() == 1)
+         const auto waveTrack = dynamic_cast<WaveTrack*>(newTracks[0].get());
+         // Also check that the track has a clip, as protection against empty
+         // file import.
+         if (waveTrack && !waveTrack->GetClipInterfaces().empty())
             resultingReader.reset(new ClipMirAudioReader {
                std::move(acidTags), fileName.ToStdString(),
-               **waveTracks.begin() });
+               *waveTrack });
       }
 
       if (addToHistory) {
@@ -1616,7 +1662,7 @@ void ProjectFileManager::Compact()
    const auto greatest = std::max<size_t>(savedState, currentState);
    std::vector<const TrackList*> trackLists;
    auto fn = [&](const UndoStackElem& elem) {
-      if (auto pTracks = TrackList::FindUndoTracks(elem))
+      if (auto pTracks = UndoTracks::Find(elem))
          trackLists.push_back(pTracks);
    };
    undoManager.VisitStates(fn, least, 1 + least);
